﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.Linq;
using DocumentFormat.OpenXml.Packaging;

namespace OpenXmlPowerTools
{
    public class RunSplitter
    {
        private static object RunSplitterTransform(XNode node)
        {
            XElement element = node as XElement;
            if (element != null)
            {
                if (element.Elements(W.r).Any())
                {
                    XElement paragraphWithSplitRuns = new XElement(W.p,
                        element.Attributes(),
                        element.Nodes().Select(n => RunSplitterTransform(n)));
                    return paragraphWithSplitRuns;
                }
                if (element.Name == W.r && element.Elements(W.t).Any())
                {
                    var collectionOfRuns = element.Elements()
                        .Where(e => e.Name != W.rPr)
                        .Select(e =>
                        {
                            if (e.Name == W.t)
                            {
                                string s = (string)e;
                                IEnumerable<XElement> collectionOfSubRuns = s.Select(c =>
                                {
                                    XElement newRun = new XElement(W.r,
                                        element.Elements(W.rPr),
                                        new XElement(W.t,
                                            c == ' ' ?
                                            new XAttribute(XNamespace.Xml + "space", "preserve") :
                                            null, c));
                                    return newRun;
                                });
                                return (object)collectionOfSubRuns;
                            }
                            else
                            {
                                XElement newRun = new XElement(W.r,
                                    element.Elements(W.rPr),
                                    e);
                                return newRun;
                            }
                        });
                    return collectionOfRuns;
                }
                return new XElement(element.Name,
                    element.Attributes(),
                    element.Nodes().Select(n => RunSplitterTransform(n)));
            }
            return node;
        }

        public static XElement SplitRuns(XElement paragraph)
        {
            XElement newParagraph = (XElement)RunSplitterTransform(paragraph);
            paragraph.ReplaceWith(newParagraph);
            return newParagraph;
        }

        private static object RunCoalescerTransform(XNode node)
        {
            XElement element = node as XElement;
            if (element != null)
            {
                if (element.Elements(W.r).Any())
                {
                    var groupedAdjacentRunsWithIdenticalFormatting =
                        element
                        .Elements()
                        .GroupAdjacent(ce =>
                        {
                            if (ce.Name != W.r)
                                return "DontConsolidate";
                            if (ce.Elements().Where(e => e.Name != W.rPr).Count() != 1 ||
                                ce.Element(W.t) == null)
                                return "DontConsolidate";
                            if (ce.Element(W.rPr) == null)
                                return "";
                            return ce.Element(W.rPr).ToString(SaveOptions.None);
                        });
                    XElement paragraphWithCoalescedRuns = new XElement(W.p,
                        groupedAdjacentRunsWithIdenticalFormatting.Select(g =>
                        {
                            if (g.Key == "DontConsolidate")
                                return (object)g;
                            string textValue = g.Select(r => r.Element(W.t).Value).StringConcatenate();
                            XAttribute xs = null;
                            if (textValue[0] == ' ' || textValue[textValue.Length - 1] == ' ')
                                xs = new XAttribute(XNamespace.Xml + "space", "preserve");
                            return new XElement(W.r,
                                g.First().Elements(W.rPr),
                                new XElement(W.t, xs, textValue));
                        }));
                    return paragraphWithCoalescedRuns;
                }
                return new XElement(element.Name,
                    element.Attributes(),
                    element.Nodes().Select(n => RunSplitterTransform(n)));
            }
            return node;
        }

        public static XElement CoalesceRuns(XElement paragraph)
        {
            XElement newParagraph = (XElement)RunCoalescerTransform(paragraph);
            paragraph.ReplaceWith(newParagraph);
            return newParagraph;
        }
    }
}
